/* -*-C++-*-
 * ###################################################################
 *  Cpptcl - connecting C++ with Tcl
 * 
 *  FILE: "member_matrix.cc"
 *                                    created: 21/10/97 {2:15:28 pm} 
 *                                last update: 20/4/98 {6:03:15 pm} 
 *  Author: Vince Darley
 *  E-mail: <darley@fas.harvard.edu>
 *    mail: Division of Engineering and Applied Sciences, Harvard University
 *          Oxford Street, Cambridge MA 02138, USA
 *     www: <http://www.fas.harvard.edu/~darley/>
 *  
 * ========================================================================
 *               Copyright (c) 1997 Vince Darley
 * ========================================================================
 *  See header file for further information
 * ###################################################################
 */

#include "member_matrix.h"

Cpptcl_IBaseClass(mem_matrix,"matrixmember",cpp_config_mem);

void mem_matrix::conversion_error(tcl_args& arg) const {
	tcl_obj obuf;
	if(read_only()) {
		obuf << " - item is read-only";
	} else {
		obuf << " to type '" << data_type() << "'";
	}			
	arg.set_conversion_type(obuf);
	arg.syntax("value");
}

void mem_matrix::get_syntax(tcl_args& arg) const {
	arg("'matrix name or explicit value'","set this matrix member to the given value");
}

int mem_matrix::parse_meta_commands(tcl_obj& tcl_, tcl_args& arg) {
	return cpp_config_mem::parse_meta_commands(tcl_,arg);
}


int wrapped_matrix::output(int dim, int offs, int nargs, tcl_obj& obj) const {
	if (dim < dimensions()) {
		for (int i = 0; i < nargs; i++) {
			int newoffs = (dim > 0 ? offs*dimension(dim-1) + i : 0);
			tcl_obj temp;
			output(dim + 1,newoffs,dimension(dim),temp);
			obj << temp << lappend;
		}
		return TCL_OK;
	}
	int newoffs = offs * dimension(dim - 1);
	get_range(newoffs,nargs,obj);
	return TCL_OK;
}


int wrapped_matrix::initialise(int dim, int offs, int nargs, const Tcl_Obj** obj) {
	if (dim < dimensions()) {
		for (int i = 0; i < nargs; i++) {
			int split_list_len;
			Tcl_Obj** split_list;
			if (Tcl_ListObjGetElements(NULL, (Tcl_Obj*)obj[i], 
									&split_list_len, &split_list) != TCL_OK) {
				return TCL_ERROR;
			}
			if (split_list_len > dimension(dim)) {
				return TCL_ERROR;
			}
			int newoffs = (dim > 0 ? offs*dimension(dim-1) + i : 0);
			if(initialise(dim + 1,newoffs,split_list_len,split_list) != TCL_OK) {
				return TCL_ERROR;
			}
		}
		return TCL_OK;
	}
	int newoffs = offs * dimension(dim - 1);
	if(set_range(newoffs,nargs,obj) != TCL_OK) {
		return TCL_ERROR;
	}
	return TCL_OK;
}

template <class T>
int wrapped_matrix_t<T>::get_range(int newoffs, int nargs, tcl_obj& obj) const {
	const T *iptr = ptr() + newoffs;
	for (int i = 0; i < nargs; i++) {
		obj << iptr[i] << lappend;
	}
	return TCL_OK;
}

int wrapped_matrix_t<int>::set_range(int newoffs, int nargs, const Tcl_Obj** obj) {
	int *iptr = ptr() + newoffs;
	for (int i = 0; i < nargs; i++) {
		if(Tcl_GetIntFromObj(NULL, (Tcl_Obj*)obj[i], iptr+i) != TCL_OK) {
			return TCL_ERROR;
		}
	}
	return TCL_OK;
}
int wrapped_matrix_t<bool>::set_range(int newoffs, int nargs, const Tcl_Obj** obj) {
	bool *iptr = ptr() + newoffs;
	for (int i = 0; i < nargs; i++) {
		int temp;
		if(Tcl_GetBooleanFromObj(NULL, (Tcl_Obj*)obj[i], &temp) != TCL_OK) {
			return TCL_ERROR;
		}
		iptr[i] = temp;
	}
	return TCL_OK;
}
int wrapped_matrix_t<long>::set_range(int newoffs, int nargs, const Tcl_Obj** obj) {
	long* iptr = ptr() + newoffs;
	for (int i = 0; i < nargs; i++) {
		if(Tcl_GetLongFromObj(NULL, (Tcl_Obj*)obj[i], iptr+i) != TCL_OK) {
			return TCL_ERROR;
		}
	}
	return TCL_OK;
}
int wrapped_matrix_t<double>::set_range(int newoffs, int nargs, const Tcl_Obj** obj) {
	double* fptr = ptr() + newoffs;
	for (int i = 0; i < nargs; i++) {
		if(Tcl_GetDoubleFromObj(NULL, (Tcl_Obj*)obj[i], fptr+i) != TCL_OK) {
			return TCL_ERROR;
		}
	}
	return TCL_OK;
}
int wrapped_matrix_t<float>::set_range(int newoffs, int nargs, const Tcl_Obj** obj) {
	float *fptr = ptr() + newoffs;
	for (int i = 0; i < nargs; i++) {
		double _fptr;
		if(Tcl_GetDoubleFromObj(NULL, (Tcl_Obj*)obj[i], &_fptr) != TCL_OK) {
			return TCL_ERROR;
		}
		fptr[i] = (float)_fptr;
	}
	return TCL_OK;
}

 
  cppMemMatrixType(bool);
  cppMemMatrixType(int);
  cppMemMatrixType(long);
  cppMemMatrixType(float);
  cppMemMatrixType(double);

template <class T>
void member_matrix<T>::tcl_get(tcl_object* o, tcl_obj& tcl) const {
	tcl << get(o);
}

template <class T>
void member_matrix<T>::tcl_set(tcl_object* o, tcl_args_reader& istr) const {
	if(read_only()) {
		istr.setstate(tcl_args_reader::Fail);
		return;
	}
	a_matrix<T> t;
	istr >> t;
	if(istr.ok()) {
		set(o,t);
	}
}

template class member_matrix<bool>;
template class member_matrix<int>;
template class member_matrix<float>;
template class member_matrix<long>;
template class member_matrix<double>;

template class wrapped_matrix_t<bool>;
template class wrapped_matrix_t<int>;
template class wrapped_matrix_t<float>;
template class wrapped_matrix_t<long>;
template class wrapped_matrix_t<double>;

